#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "mysqlstuff.h" /* Currently assumes a MYSQL database */

#define GEQVERSION "2.07"
#define GEQUELVERSION " v" GEQVERSION ", July 2019 "

#define LIBGLADEXMLFILE "gequel.glade-3"

#define DEBUG /* MACRO to make debugging easier */
#define DEBUGLEVEL 1 /* only for ldebug calls */

enum Tabs   { RECORDSTAB, FIELDSTAB, SEARCHTAB, QUERYTAB, TABCOUNT };

#define STDFONT  "Sans 10"
#define BOLDFONT "Sans Bold 11"
#define REDCOLOR { 0.8, 0.2, 0.1 }
    
#define FONTWIDTH  7 /* reserved space for one character in browser column */

#define SELECT  "SELECT * FROM %s "
#define INSERT  "INSERT INTO %s VALUES ("
#define UPDATE  "UPDATE %s SET "
#define DELETE  "DELETE FROM %s "
#define WHERE   " WHERE "
#define EQUAL   " = "
#define LIKE    " LIKE "
#define AND     " AND "
#define ORDERBY " ORDER BY "

#define PASSIVE  0
#define ACTIVE   1
#define KEYFIELD 2

#define COUNT    0
#define NOCOUNT  1

#define MAXLOGSIZE 500000
#define MAXQUERY   10000
#define MAXMESSAGE 2048
#define MAXSTRING  256

#define DELIMITER ","         /* field delimiter for ASCII output. Here set to comma */
/* #define DELIMITER "\t" */  /* field delimiter for ASCII output. Here set to <TAB> */

#define DEFSOCKETNAME "/var/run/mysqld/mysqld.sock"

/* contains info for one single record */
typedef struct _Record
{
   char      *table;        /* currently selected table */
   int       numflds;       /* number of fields (without opt. indices) */
   int       sortkey;       /* number of field we'll use to sort */
   SQL       *dbsock;       /* MYSQL style : int dbsock for MSQL */
   SQL_FIELD *fields;       /* stores fields for current table */
   char      *selectedname; /* hook for fileselector to store choice */

   /* until here we overlap with the Tabinfo struct : consider using union ? */
   
   char      **contents;    /* field contents as obtained from database */
   GtkWindow *window;       /* top window handle for this record */
   GtkWidget *gtable;       /* holds info: do not confuse with dbase table ! */
   GtkEntry  **entries;     /* (changed) field contents inside GUI textentry */
} Record;

/* contains info for one single browser panel (tab) */
typedef struct _TabInfo
{
   char      *table;        /* currently selected table */
   int       numflds;       /* number of dbfields (without opt. indices) */
   int       sortkey;       /* number of field we'll use to sort */
   SQL       *dbsock;       /* MYSQL style : int dbsock for MSQL */
   SQL_FIELD *fields;       /* stores fields for current table */
   char      *selectedname; /* hook for fileselector to store choice */

   /* until here we overlap with the Tabinfo struct : consider using union ? */

   int        selectedrow; /* number of selected row (path of 1 element) */

   int          nrcols;    /* number of columns (often fields) in browser */
   int          nrrows;    /* number of rows (records) in browser */
   
   gchar        **titles;  /* titles to put above columns */
   
   GtkTreeView  *view;     /* treeview widget in notebook serving as viewer for list */
   GtkListStore *list;     /* pointer to list model as show in treeview widget */
   GtkCellRenderer **renderer; /* for changing color and font in the treeview widget */
   
   GtkCList     *clist;    /* DUMMY, should be removed !!! */
   
   gboolean     inuse;     /* TRUE if tab contains content */
   gboolean     refresh;   /* TRUE if tab needs a refresh */
   int          index;     /* Tab number in notebook */
} TabInfo;

/* contains all global application info */
typedef struct _Params
{
   char      table[MAXSTRING]; /* currently selected table */
   int       numflds;          /* number of fields (only for counter) */
   int       sortkey;          /* number of field we'll use to sort */
   SQL       *dbsock;          /* MYSQL style : int dbsock for MSQL */
   SQL_FIELD *fields;          /* stores fields for current table */

   GtkWindow *mainwindow;
   GtkWidget *querywindow;
   GtkWidget *logwindow;       /* widget main window: need it for log messages */
   GtkWidget *notebook;

   /* Gtk2 embeds text buffer widgets inside a GtkTextView widget, while Gtk1.2
    uses a single GtkText widget for both parts */
  
   GtkTextBuffer *querybuffer;
   GtkTextBuffer *logbuffer; 
   
   int       numrecs;      /* number of records (only for counter) */

   char      *user;
   char      *pwd;
   char      *socketname;
   char      *server;
   char      database[MAXSTRING];
   char      *selectedname; /* hook for fileselector to store choice */
   int       countflag;     /* count records when loading new table ? */
   
   int       nrtabs;        /* number of tabpages in browser notebook */
   TabInfo   tabinfo[TABCOUNT];
   
   GtkOptionMenu *tablelist;
   GtkOptionMenu *dbaselist;

} Params;

/* main.c : */
Params    *getParams(void);    /* returns the Params structure from main */
GtkWindow *getTopWindow(void); /* returns the top_window id as stored in main */
GtkWindow* create_mainwindow (void);
GtkWindow* create_recordwindow (void);
GtkWindow* create_dialogbox (void);

/* startup.c */
void gequelInit(int argc, char *argv[]);
void storeWidgets(GtkWindow *main);
void connectMainWindowSignals(GtkWindow *main);

/* dialog.c : */
void ldebug(int level, char *format, ...);
void defaultAction(GtkButton *button, gpointer dummy);
void showMessageBox(char *s);
void showConfirmBox(char *s, gpointer yes, gpointer no, gpointer hook);
void showVarArgMessage(char *s,...);
void showMessage(char *s);
void setLogWindow(void);
void showSqlError(SQL *sock, char *s,...);
void writeLog(char *s,...);
void saveLogAction(GtkWidget *w, gpointer user_data );

/* helpers.c : */
void setStringValue(GtkWindow *window, char *widgetname,char *string);
int  executeQuery(SQL *sock, char *query);
int  getNumRecords(void);
void setGUICounters(int nrFields, int nrRecords);
SQL_FIELD *copyFields(SQL_FIELD *src, int numflds);
void      freeFields(SQL_FIELD *tgt, int numflds);
void on_notebooktab_switched
  (GtkNotebook *,GtkNotebookPage *page, int pagenum, gpointer user_data);
void setBoldRedFont(GtkWidget *widget);
void setBoldFont(GtkWidget *widget);
  
/* sqlconnect.c : */
void setNewServer(char *server, char *socketname);
void setDBaseList(char *preselect);
void setTableList(char *preselect);
void setNewDatabase(char *database);
void setNewTable(char *table);

/* optionactions.c : */
void on_database_activate(GtkMenuItem *item, gpointer  user_data);
void on_table_activate(GtkMenuItem *item, gpointer  user_data);

/* showrecords.c : */
void record_doubleclick(GtkTreeView       *tree_view,
			GtkTreePath       *path,
			GtkTreeViewColumn *column,
			gpointer           data);
void showRecords(void);

/* showfields.c : */
void field_doubleclick(GtkTreeView       *tree_view,
		       GtkTreePath       *path,
		       GtkTreeViewColumn *column,
		       gpointer           data);
void setSortKey(int row);
int  getSortKey(void);
void showFields(void);

/* showquery.c : */
void qrecord_doubleclick(GtkTreeView       *tree_view,
			 GtkTreePath       *path,
			 GtkTreeViewColumn *column,
			 gpointer           data);
void showQuery(char *query);
void loadQueryAction(GtkWidget *w, gpointer user_data );
void saveQueryAction(GtkWidget *w, gpointer user_data );

/* showsearch.c : */
void srecord_doubleclick(GtkTreeView       *tree_view,
			 GtkTreePath       *path,
			 GtkTreeViewColumn *column,
			 gpointer           data);
void showSearch(Record *rec);

/* recordform.c : */
void on_recorddelete_clicked(GtkButton *dummy, gpointer rec); /* callback */
void destroyRecordWindow(Record *rec);
void newModifyWindow(TabInfo *ti);
void newCopyRecordWindow(Record *rec);
void newCopyMainWindow(TabInfo *ti);
void newInsertWindow(void);
void newSearchWindow(void);
void saveRecordAction(GtkWidget *w, gpointer user_data);

/* sqlactions.c : */
char *addStringQuotes(int type);
void updateRecordAction(GtkButton *button, gpointer rec);
void deleteRecordAction(GtkButton *button, gpointer rec);
void insertRecordAction(GtkButton *button, gpointer rec);

/* recordhandlers.c : */
Record  *newRecordstruct(TabInfo *ti);
Record  *copyRecordstruct(Record *rec);
void    destroyRecordstruct(Record *);

/* tabhandlers.c : */
void    view_selected(GtkTreeSelection *selection, gpointer tabid);
void    forceRefreshRecordTab(void); // always rereshes the tab
void    refreshRecordTab(void); // only refreshes when autofresh is on
void    refreshFieldTab(void);
void    setTabColumnWidths(TabInfo *ti, int tabid);
void    saveTabAction(GtkWidget *w, gpointer user_data);
void    initTabList(TabInfo *ti,int tabid);
TabInfo *initTabInfo(int tabid, int nrcols, int nrrows);
void    clearTabList(TabInfo *ti);
void    clearTabInfo(TabInfo *ti);
void    clearAllTabs(void);

/* filestuff.c : */
void newFileSelector(char *preset, char **result, GtkFileChooserAction action);

/* libglade.c : */
void      add_to_search_path(const gchar *dir);
GtkWidget *lookup_widget(GtkWindow *root, char *widgetname);
GtkWidget *create_widget(char *widgetname);  
